@@ -7,7 +7,9 @@ import configparser
 import datetime
 import hashlib
 import io
+import json
 import logging
+import threading
 import time
 import weakref
 from typing import Any
@@ -47,6 +49,9 @@ class AbstractClient(abc.ABC):
     # Empty dictionary, to be overridden by the subclass
     HOSTED_URLS = {}
 
+    # Login lock, re-entrant because the first request requires another login request
+    LOGIN_LOCK = threading.RLock()
+
     @classmethod
     def _get_login_params(
         cls, conf: configparser.ConfigParser, section_name: str
@@ -317,8 +322,10 @@ class AbstractClient(abc.ABC):
         :raises CommunicationError: if there was an error connecting to the resource
         :raises ApiError: if there was on error on the server side
         """
-        if not self._is_logged_in():
-            self._login()
+        # Login in a 'reentrant' way, and play nice with multiple threads
+        with self.LOGIN_LOCK:
+            if not self._is_logged_in():
+                self._login()
 
         try:
             response = self._session.request(
@@ -398,6 +405,159 @@ class PortalClient(AbstractClient):
             self._session = requests.sessions.session()
         self.post("login", function=None, data=self._login_params)
 
+    def get_monitored_hosts(self, breach_uuid, start_time, end_time, customer):
+        """
+        Get the list of monitored hosts.
+
+        :param str breach_uuid: the breach uuid
+        :param str start_time: the start time of the breach as returned by 'get_breach'
+        :param str end_time: the end time of the breach as returned by 'get_breach'
+        :param str|None customer: the customer username if different from the current one
+        :rtype: list[dict]
+        :return: a dictionary containing hosts information
+        """
+        params = tau_clients.purge_none(
+            {
+                "customer": customer,
+                "breach_uuid": breach_uuid,
+                "start_time": start_time,
+                "end_time": end_time,
+            }
+        )
+        return self.get("net", "monitored_host/all/list", params=params)
+
+    def get_breach(self, breach_uuid: str, customer: Optional[str] = None) -> Dict:
+        """
+        Get breach information.
+
+        :param str breach_uuid: the breach uuid
+        :param str|None customer: the customer username if different from the current one
+        :rtype: dict
+        :return: a dictionary containing breach information
+        """
+        params = tau_clients.purge_none(
+            {
+                "customer": customer,
+                "breach_uuid": breach_uuid,
+            }
+        )
+        return self.get("net", "breach/get", params=params)
+
+    def get_breach_summary(
+        self,
+        breach_uuid: str,
+        start_time: str,
+        end_time: str,
+        customer: Optional[str] = None,
+    ) -> List[Dict]:
+        """
+        Get breach summary split in phases.
+
+        :param str breach_uuid: the breach uuid
+        :param str start_time: the start time of the breach as returned by 'get_breach'
+        :param str end_time: the end time of the breach as returned by 'get_breach'
+        :param str|None customer: the customer username if different from the current one
+        :rtype: list[dict]
+        :return: a dictionary containing breach information
+        """
+        params = tau_clients.purge_none(
+            {
+                "customer": customer,
+                "breach_uuids": breach_uuid,
+                "start_time": start_time,
+                "end_time": end_time,
+            }
+        )
+        return self.get("net", "breach/phases_summary", params=params)
+
+    def get_breach_phases(
+        self,
+        breach_uuid: str,
+        start_time: str,
+        end_time: str,
+        customer: Optional[str] = None,
+    ) -> List[Dict]:
+        """
+        Get breach phases in detail.
+
+        :param str breach_uuid: the breach uuid
+        :param str start_time: the start time of the breach as returned by 'get_breach'
+        :param str end_time: the end time of the breach as returned by 'get_breach'
+        :param str|None customer: the customer username if different from the current one
+        :rtype: list[dict]
+        :return: a dictionary containing breach phases
+        """
+        params = tau_clients.purge_none(
+            {
+                "customer": customer,
+                "breach_uuid": breach_uuid,
+                "start_time": start_time,
+                "end_time": end_time,
+            }
+        )
+        return self.get("net", "breach/phases", params=params)
+
+    def get_breach_evidence(
+        self,
+        breach_uuid: str,
+        include_events: bool = False,
+        customer: Optional[str] = None,
+    ) -> List[Dict]:
+        """
+        Get breach evidences.
+
+        :param str breach_uuid: the breach uuid
+        :param bool include_events: whether to return reference events for each evidence
+        :param str|None customer: the customer username if different from the current one
+        :rtype: list[dict]
+        :return: a list of dictionaries containing evidences
+        """
+        params = tau_clients.purge_none(
+            {
+                "customer": customer,
+                "breach_uuid": breach_uuid,
+                "extended": include_events,
+            }
+        )
+        return self.get("net", "breach/evidence", params=params)
+
+    def list_events(
+        self,
+        start_time: str,
+        end_time: str,
+        max_results: Optional[int] = None,
+        offset_results: Optional[int] = None,
+        event_outcome: Optional[str] = None,
+        customer: Optional[str] = None,
+        user_id: Optional[int] = None,
+    ) -> List[Dict[str, Any]]:
+        """
+        List all events.
+
+        :param str start_time: the start time
+        :param str end_time: the end time
+        :param int|None max_results: the maximum result to return (useful when paginating)
+        :param int|None offset_results: the offset when paginating
+        :param str|None event_outcome: the outcome of the event (DETECTION/INFO/ALL)
+        :param str|None customer: the customer username if different from the current one
+        :param int|None user_id: the user id (requires special privileges)
+        :rtype: list[dict[str, any]]
+        :return: the list of events
+        """
+        params = tau_clients.purge_none(
+            {
+                "customer": customer,
+                "user_id": user_id,
+                "start_time": start_time,
+                "end_time": end_time,
+                "max_results": max_results,
+                "offset_results": offset_results,
+                "event_outcome": event_outcome,
+                "orderby": "event_end_time DESC",
+            }
+        )
+        return self.get("net", "event/list", params=params)
+
     def get_event(
         self,
         event_id: str,
@@ -553,6 +713,7 @@ class PortalClient(AbstractClient):
         referer: Optional[str] = None,
         user_agent: Optional[str] = None,
         bypass_cache: bool = False,
+        delete_after_analysis: bool = False,
     ) -> Dict[str, Any]:
         """
         Upload a URL to be analyzed.
@@ -561,6 +722,7 @@ class PortalClient(AbstractClient):
         :param str|None referer: the referer
         :param str|None user_agent: the user agent
         :param bool bypass_cache: bypass_cache
+        :param bool delete_after_analysis: whether to delete after analysis
         :rtype: dict[str, any]
         :return: a dictionary like:
             {
@@ -589,6 +751,7 @@ class PortalClient(AbstractClient):
                 "bypass_cache": bypass_cache,
                 "referer": referer,
                 "user_agent": user_agent,
+                "delete_after_analysis": delete_after_analysis,
             }
         )
         return self.post("analysis", "submit_url", data=data)
@@ -601,6 +764,8 @@ class PortalClient(AbstractClient):
         analysis_env: Optional[str] = None,
         allow_network_traffic: bool = True,
         bypass_cache: bool = False,
+        bypass_prefilter: bool = False,
+        delete_after_analysis: bool = False,
     ) -> Dict[str, Any]:
         """
         Upload a file to be analyzed.
@@ -611,6 +776,8 @@ class PortalClient(AbstractClient):
         :param str|None analysis_env: if set, e.g windowsxp
         :param bool allow_network_traffic: if set to False, deny network connections
         :param bool bypass_cache: whether to re-process a file (requires special permissions)
+        :param bool bypass_prefilter: whether to skip the prefilter (requires special permissions)
+        :param bool delete_after_analysis: whether to delete after analysis
         :rtype: dict[str, any]
         :return: a dictionary like:
             {
@@ -640,6 +807,8 @@ class PortalClient(AbstractClient):
                 "analysis_env": analysis_env,
                 "allow_network_traffic": allow_network_traffic,
                 "bypass_cache": bypass_cache,
+                "bypass_prefilter": bypass_prefilter,
+                "delete_after_analysis": delete_after_analysis,
             }
         )
         files = {"file": (file_name, file_data, "application/octet-stream")}
@@ -655,6 +824,8 @@ class AnalysisClient(AbstractClient):
 
     HOSTED_URLS = tau_clients.NSX_DEFENDER_ANALYSIS_URLS
 
+    DELETE_DATE_FMT = "%Y-%m-%d %H:%M:%S.%f"
+
     def _login(self) -> None:
         """Implement interface (analysis client relies on 'key' and 'api_token')."""
         if self._session is None:
@@ -808,6 +979,37 @@ class AnalysisClient(AbstractClient):
         )
         return self.get("analysis", "completed", params=params)
 
+    def get_pending(
+        self,
+        after: Optional[Union[datetime.datetime, str]] = None,
+        before: Optional[Union[datetime.datetime, str]] = None,
+    ) -> Dict[str, Any]:
+        """
+        Get the list of uuids of tasks that were completed within a given time frame.
+
+        :param datetime.datetime|str|None after: request tasks completed after this time
+        :param datetime.datetime|str|None before: request tasks completed before this time
+        :return: a dictionary like:
+            {
+                'tasks': ['182a645e7020001000de1474baf8b7b9'],
+                'after': '2021-07-14 09:10:00',
+                'more_results_available': 0,
+                'resume': '2021-07-14 14:09:47',
+                'before': '2021-07-14 14:09:59'
+            }
+        """
+        if hasattr(before, "strftime"):
+            before = before.strftime(tau_clients.DATETIME_FMT)
+        if hasattr(after, "strftime"):
+            after = after.strftime(tau_clients.DATETIME_FMT)
+        params = tau_clients.purge_none(
+            {
+                "before": before,
+                "after": after,
+            }
+        )
+        return self.get("analysis", "get_pending", params=params)
+
     def get_task_metadata(self, uuid: str) -> Dict[str, str]:
         """
         Get the metadata of a given task.
@@ -984,6 +1186,7 @@ class AnalysisClient(AbstractClient):
                     "report_uuid": "4ffa644851ebe118wu1pbwlt3C44cwAYRtfdXY5L3gRhHJ_QANAyuQ",
                     "artifact_name": "process_snapshots_2",
                     "artifact_type": "process_snapshot",
+                    "delete_date": optional datetime object,
                 },
                 ...
             ]
@@ -996,15 +1199,33 @@ class AnalysisClient(AbstractClient):
         result = self.get_result(uuid, include_report=False)
         for report in result.get("reports", []):
             if not report_types or any(x in report["report_versions"] for x in report_types):
-                temp = self.get_result(uuid, report_uuid=report["report_uuid"])
-                for item in temp.get("report", {}).get("analysis_metadata", []):
-                    if not metadata_types or item["metadata_type"] in metadata_types:
+                ret.append(
+                    {
+                        "task_uuid": uuid,
+                        "report_uuid": report["report_uuid"],
+                        "artifact_name": "report",
+                        "artifact_type": tau_clients.METADATA_TYPE_REPORT,
+                        "delete_date": None,
+                    }
+                )
+                report_data = self.get_result(uuid, report_uuid=report["report_uuid"])
+                metadata_items = report_data.get("report", {}).get("analysis_metadata", [])
+                for item in metadata_items:
+                    if item["metadata_type"] in metadata_types or not metadata_types:
+                        try:
+                            delete_date = datetime.datetime.strptime(
+                                item.get("delete_date") or item.get("retention_date"),
+                                self.DELETE_DATE_FMT,
+                            )
+                        except TypeError:
+                            delete_date = None
                         ret.append(
                             {
                                 "task_uuid": uuid,
                                 "report_uuid": report["report_uuid"],
                                 "artifact_name": item["name"],
                                 "artifact_type": item["metadata_type"],
+                                "delete_date": delete_date,
                             }
                         )
         return ret
@@ -1026,15 +1247,23 @@ class AnalysisClient(AbstractClient):
         :rtype: NamedBytesIO
         :return: the artifact
         """
-        params = tau_clients.purge_none(
-            {
-                "uuid": uuid,
-                "report_uuid": report_uuid,
-                "artifact_name": artifact_name,
-                "allow_datacenter_redirect": allow_datacenter_redirect,
-            }
-        )
-        return self.get("analysis", "get_report_artifact", params=params, fmt=None, raw=True)
+        if artifact_name == "report":
+            report = self.get_result(uuid, report_uuid, include_report=True)
+            return tau_clients.NamedBytesIO(
+                content=json.dumps(report, indent=2).encode("utf-8"),
+                name="report.json",
+                mime_type="application/json",
+            )
+        else:
+            params = tau_clients.purge_none(
+                {
+                    "uuid": uuid,
+                    "report_uuid": report_uuid,
+                    "artifact_name": artifact_name,
+                    "allow_datacenter_redirect": allow_datacenter_redirect,
+                }
+            )
+            return self.get("analysis", "get_report_artifact", params=params, fmt=None, raw=True)
 
     def submit_file(
         self,
@@ -1044,7 +1273,9 @@ class AnalysisClient(AbstractClient):
         analysis_env: Optional[str] = None,
         allow_network_traffic: bool = True,
         bypass_cache: bool = False,
+        bypass_prefilter: bool = False,
         include_report: bool = False,
+        delete_after_analysis: bool = False,
     ) -> Dict[str, Any]:
         """
         Upload a file to be analyzed.
@@ -1054,8 +1285,10 @@ class AnalysisClient(AbstractClient):
         :param str|None password: if set, use it to extract the sample
         :param str|None analysis_env: if set, e.g windowsxp
         :param bool allow_network_traffic: if set to False, deny network connections
-        :param bool bypass_cache: whether to re-process a file
+        :param bool bypass_cache: whether to re-process a file (requires special permissions)
+        :param bool bypass_prefilter: whether to skip the prefilter (requires special permissions)
         :param bool include_report: whether to include the report in the result
+        :param bool delete_after_analysis: whether to delete after analysis
         :rtype: dict[str, any]
         :return: a dictionary in the following form if the analysis is already available:
             {
@@ -1081,8 +1314,10 @@ class AnalysisClient(AbstractClient):
         data = tau_clients.purge_none(
             {
                 "bypass_cache": bypass_cache and 1 or None,
+                "bypass_prefilter": bypass_prefilter and 1 or None,
                 "analysis_env": analysis_env,
                 "allow_network_traffic": allow_network_traffic and 1 or None,
+                "delete_after_analysis": delete_after_analysis and 1 or None,
                 "filename": file_name,
                 "password": password,
                 "full_report_score": -1,
@@ -1104,6 +1339,7 @@ class AnalysisClient(AbstractClient):
         user_agent: Optional[str] = None,
         bypass_cache: bool = False,
         include_report: bool = False,
+        delete_after_analysis: bool = False,
     ) -> Dict[str, Any]:
         """
         Upload an URL to be analyzed.
@@ -1111,8 +1347,9 @@ class AnalysisClient(AbstractClient):
         :param str url: the url to analyze
         :param str|None referer: the referer
         :param str|None user_agent: the user agent
-        :param bool bypass_cache: bypass_cache
+        :param bool bypass_cache: bypass_cache (requires special permissions)
         :param bool include_report: whether to include the report in the result
+        :param bool delete_after_analysis: whether to delete after analysis
         :rtype: dict[str, any]
         :return: a dictionary like the following if the analysis is already available:
             {
@@ -1141,6 +1378,7 @@ class AnalysisClient(AbstractClient):
                 "referer": referer,
                 "bypass_cache": bypass_cache and 1 or None,
                 "user_agent": user_agent or None,
+                "delete_after_analysis": delete_after_analysis or None,
             }
         )
         ret_data = self.post("analysis", ["submit", "url"], data=data)
